/* annotation-builder.c:
 *
 * vim:smartindent ts=8:sts=2:sta:et:ai:shiftwidth=2
 ****************************************************************
 * Copyright (C) 2005 Canonical Limited
 *        Authors: Robert Collins <robert.collins@canonical.com>
 *
 * See the file "COPYING" for further information about
 * the copyright and warranty status of this work.
 */


#include "hackerlab/bugs/panic.h"
#include "hackerlab/char/str.h"
#include "libarch/annotation-builder.h"

/* local static functions */
static int
arch_annotation_builder_find_file (arch_annotation_builder_t *builder, t_uchar const *file_id)
{
    int position;
    for (position = 0; position < ar_size_arch_annotated_file (builder->files); ++position)
        if (!str_cmp (file_id, builder->files[position]->file_id))
            return position;
    return -1;
}

void
arch_annotation_builder_init (arch_annotation_builder_t *builder)
{
    builder->files = NULL;
    builder->current_file = NULL;
    builder->current_patch = NULL;
}

void 
arch_annotation_builder_finalise (arch_annotation_builder_t *builder)
{
    int position;
    for (position = 0; position < ar_size_arch_annotated_file (builder->files); ++position)
        arch_annotated_file_delete (&builder->files[position]);
    ar_free_arch_annotated_file (&builder->files);
    builder->current_file = NULL;
    arch_patch_delete (&builder->current_patch);
}

/**
 * \@brief note a file by its observed id, and the number of lines we need to annotate.
 */
void
arch_annotation_builder_add_file (arch_annotation_builder_t *builder, t_uchar const *file_id, int size_if_start)
{
    int position = arch_annotation_builder_find_file (builder, file_id);
    
    if (position != -1)
      {
        builder->current_file = builder->files[position];
        return;
      }
    invariant (size_if_start != -1);
    (*ar_ref_arch_annotated_file (&builder->files, ar_size_arch_annotated_file (builder->files))) = arch_annotated_file_new (file_id, size_if_start);
    builder->current_file = builder->files[ar_size_arch_annotated_file (builder->files) - 1];
}

int
arch_annotation_builder_has_file (arch_annotation_builder_t *builder, t_uchar const *file_id)
{
    return arch_annotation_builder_find_file(builder, file_id) != -1;
}

arch_annotated_file_t *
arch_annotation_builder_get_file (arch_annotation_builder_t *builder, t_uchar const *file_id)
{
    arch_annotated_file_t *result;
    int position = arch_annotation_builder_find_file (builder, file_id);
    if (position == -1)
        return NULL;
    result = *ar_ref_arch_annotated_file (&builder->files, position);
    for (; position < ar_size_arch_annotated_file (builder->files) - 1; ++position )
      {
        *ar_ref_arch_annotated_file (&builder->files, position) = *ar_ref_arch_annotated_file (&builder->files, position + 1);
      }
    ar_pop_arch_annotated_file (&builder->files);
    builder->current_file = NULL;
    return result;
}

/**
 * \@brief add a patch 
 */
void 
arch_annotation_builder_add_patch (arch_annotation_builder_t *builder, arch_patch_id *patch, t_uchar const * creator)
{
    arch_patch_delete (&builder->current_patch);
    builder->current_patch = arch_patch_new (patch->patch_id);
    arch_patch_creator_set (builder->current_patch, creator);
}

void
arch_annotation_builder_subtract_line (arch_annotation_builder_t *builder, int old_offset)
{
    invariant (builder->current_file != NULL);
    invariant (builder->current_patch != NULL);
    arch_annotated_file_note_subtract (builder->current_file, old_offset);
}

void
arch_annotation_builder_add_line (arch_annotation_builder_t *builder, int line)
{
    invariant (builder->current_file != NULL);
    invariant (builder->current_patch != NULL);
    arch_annotated_file_note_line (builder->current_file, line, builder->current_patch, 0);

}

void
arch_annotation_builder_process_changes(arch_annotation_builder_t *builder, patch_line_changes_list changes)
{
    int position;
    for (position = 0; position < ar_size_patch_line_change (changes); ++ position)
      {
        if (changes[position].operation == PATCH_ADD_LINE)
          {
            arch_annotation_builder_add_line(builder, changes[position].offset);
          }
        else
          {
            arch_annotation_builder_subtract_line(builder, changes[position].offset);
          }
      }
}
