/* build-revision.c:
 *
 * vim:smartindent ts=8:sts=2:sta:et:ai:shiftwidth=2
 ****************************************************************
 * Copyright (C) 2003 Tom Lord
 *
 * See the file "COPYING" for further information about
 * the copyright and warranty status of this work.
 */


#include "hackerlab/bugs/panic.h"
#include "hackerlab/mem/alloc-limits.h"
#include "hackerlab/mem/mem.h"
#include "hackerlab/char/str.h"
#include "hackerlab/fs/file-names.h"
#include "hackerlab/vu/safe.h"
#include "libfsutils/copy-file.h"
#include "libfsutils/dir-listing.h"
#include "libfsutils/tmp-files.h"
#include "libfsutils/rmrf.h"
#include "libfsutils/link-tree.h"
#include "libfsutils/ensure-dir.h"
#include "libarch/ancestry.h"
#include "libarch/arch.h"
#include "libarch/namespace.h"
#include "libarch/libraries.h"
#include "libarch/project-tree.h"
#include "libarch/archive.h"
#include "libarch/archives.h"
#include "libarch/apply-changeset.h"
#include "libarch/pristines.h"
#include "libarch/invent.h"
#include "libarch/inode-sig.h"
#include "libarch/chatter.h"
#include "libarch/patch-logs.h"
#include "libarch/library-txn.h"
#include "libarch/build-revision.h"
#include "libarch/debug.h"



static void finalize_library_revision(t_uchar const * const dest_directory,
                               t_uchar * tmp_dir, 
                               t_uchar * patch_set_dir,
                               t_uchar * archive, 
                               t_uchar const * const revision);

static int
build_revision (int * patch_list_length,
		t_uchar const * const dest_dir,
                struct arch_archive * arch,
                t_uchar const * const archive, 
		t_uchar const * const revision,
                t_uchar const * const cache_dir,
		int try_link,
                t_uchar const * const non_sparse_lib);

static void
select_patch_list (rel_table * patch_list, rel_table candidate);

static void 
make_patch_list (rel_table * patch_list,
		 t_uchar const * const ancestor_archive, 
		 t_uchar const * const ancestor_revision, 
                 t_uchar const * const descendant_archive, 
		 t_uchar const * const descendant_revision,
		 assoc_table  contin_list,
		 struct arch_archive *arch);

static int 
patch_direction (rel_table patch_list, t_uchar const *const archive, t_uchar const *const revision);

static int 
decrement_revision (struct arch_archive ** my_arch, 
                    t_uchar ** my_archive, 
		    t_uchar ** my_revision, 
		    assoc_table contin_list,
		    struct arch_archive * arch, 
		    t_uchar const * const archive, 
		    t_uchar const * const revision);

static void
find_patchlists (rel_table * loc_patches,
		 rel_table * arch_patches,
                 t_uchar const * const archive, 
		 t_uchar const *const revision,
		 struct arch_archive *arch,
		 assoc_table * contin_list,
		 t_uchar const * const cache_dir);

static int
direct_descendant (struct arch_archive * arch, 
                   t_uchar const * const revision, 
                   t_uchar const * const candidate_descendant);


static t_uchar * 
find_previous_library (t_uchar const * const archive, t_uchar const * const revision);

static int
find_multisource (t_uchar ** lib_rev, 
		  t_uchar ** arch_rev, 
		  t_uchar ** continuation, 
		  struct arch_archive * arch, 
		  t_uchar const * const revision,
		  t_uchar const * const cache_dir);

static t_uchar *
scan_archive (enum arch_revision_type *type,
              int * num_scanned,
              t_uchar ** newest_cached, 
              t_uchar ** oldest_cached, 
	      struct arch_archive *arch, 
	      t_uchar const * const start_revision, 
	      t_uchar const * const end_revision,
              int max_count_after_cacherev);

static void 
find_prev_revision (assoc_table  * contin_list,
		    t_uchar ** local_rev, 
		    t_uchar ** arch_rev, 
		    struct arch_archive * arch, 
		    t_uchar const * const start_revision,
		    t_uchar const * const cache_dir);

int find_library_forwards (t_uchar ** found_archive, t_uchar ** found_revision, t_uchar const * const archive, t_uchar const * const revision);

static int 
copy_local_revision (t_uchar const * const dest_dir,  
		     t_uchar const * const archive, 
		     t_uchar const * const revision, 
		     t_uchar const * const cache_dir,
		     int try_link);

static int 
copy_library_revision (t_uchar const * const dest_dir, 
                       t_uchar const * const archive,
                               t_uchar const * const revision, 
                               int try_link);

static int 
copy_pristine_revision (t_uchar const * const dest_dir,  
                        t_uchar const * const archive, 
                        t_uchar const * const revision, 
                        t_uchar const * const cache_dir);

static void
move_tmp_revision (t_uchar const * const dest_dir, t_uchar const * const tmpdir);

static int 
copy_archive_revision(t_uchar const * const dest_dir,
                      struct arch_archive *arch,
                      t_uchar const * const revision);

static void 
copy_cached_revision (t_uchar const * const dest_dir, 
                      struct arch_archive * arch, 
                      t_uchar const * const revision);

static void
copy_import_revision (t_uchar const * const dest_dir, 
                      struct arch_archive * arch, 
                      t_uchar const * const revision);

static t_uchar * 
full_revision (t_uchar const * const archive, t_uchar const * const revision);

static int
find_for_previous_revision (struct arch_archive ** prev_arch, 
                            t_uchar ** prev_archive, 
                            t_uchar ** prev_revision, 
                            assoc_table contin_list,
                            struct arch_archive * arch, 
                            t_uchar * archive,
                            t_uchar * revision);

static void
apply_revision_patch (t_uchar const * const dest_tree_root, 
                      struct arch_archive * arch, 
                      t_uchar const * const revision,
                      int reverse);

static void 
apply_revisions (t_uchar const * const dest_dir, 
                 struct arch_archive * arch, 
                 rel_table patch_list,
                 int apply_direction,
                 t_uchar const * const non_sparse_lib);

static void
add_to_library (t_uchar const * const tree_dir, 
                t_uchar const * const library, 
                t_uchar * archive, 
                t_uchar * revision);

/* return non-0 if it came from a library.
 */
int
arch_build_revision (t_uchar const * const dest_dir,
                     struct arch_archive * arch,
                     t_uchar const * const archive, 
                     t_uchar const * const revision,
                     t_uchar const * const cache_dir)
{
  struct arch_archive * my_arch = arch;
  int from_library = 0;
  int n_patches;
  if (my_arch == 0)
    my_arch = arch_archive_connect_branch(archive, NULL);
  invariant (my_arch != 0);

  from_library = build_revision (&n_patches, dest_dir, my_arch, archive, revision, cache_dir, 0, NULL) == 1;
  if (n_patches > 1) from_library = 0;
  if (my_arch != arch) 
    arch_archive_close (my_arch);
  return from_library;
}

void
arch_build_library_revision(int no_patch_noise, 
                            t_uchar const * const dest_directory,
                            struct arch_archive * arch, 
                            t_uchar const * const revision,
                            t_uchar const * const opt_library, 
                            t_uchar const * const opt_same_dev_path,
                            int sparse,
                            arch_apply_changeset_report_callback callback,
                            int escape_classes)
{
  t_uchar * dest_dir_dir = 0;
  t_uchar * tmp_dir = 0;
  t_uchar * my_opt_library = str_save (0, opt_library);
  int n_patches;

  dest_dir_dir = file_name_directory_file (0, dest_directory);
  ensure_directory_exists (dest_dir_dir);
  tmp_dir = tmp_file_name (dest_dir_dir, ",,new-revision");  
 
  if (my_opt_library == NULL) 
  {
    my_opt_library = arch_library_add_choice (arch->official_name, revision, my_opt_library, dest_dir_dir);
  }
 

  /* 
   * WARNING: tag patches are not reversible in older 
   * arch archives, so do not consider removing the cross-archive limit.
   */

  safe_mkdir (tmp_dir, 0777);
  build_revision (&n_patches, tmp_dir, arch, arch->official_name, revision, 0, 1, (sparse) ? NULL : my_opt_library);

  finalize_library_revision (dest_directory, tmp_dir, NULL, arch->official_name, revision);

  lim_free (0, dest_dir_dir);
  lim_free (0, tmp_dir);
  lim_free (0, my_opt_library);
}


/* \brief Turn a temp dir containing a revision into a library revision
 *
 * Finish up adding ancillary files to the tmp-dir and then
 * install it in the library.
 *
 * \param dest_directory The directory to store in
 * \param tmp_dir The directory the revision is currently in
 * \param the (official) archive of the revision
 * \param the unqualified revision name of the revision being produced
 */
void finalize_library_revision(t_uchar const * const dest_directory,
                               t_uchar * tmp_dir, 
                               t_uchar * patch_set_dir,
                               t_uchar * archive, 
                               t_uchar const * const revision)
{
  if (patch_set_dir)
    rmrf_file (patch_set_dir);
  {
    rel_table index = 0;
    t_uchar * index_by_name_path = 0;
    t_uchar * index_path = 0;
    int index_by_name_fd = -1;
    int index_fd = -1;
    arch_project_tree_t * tree;

    {
      arch_patch_id patch_id;
    
      arch_patch_id_init_archive (&patch_id, archive, revision);
      arch_set_tree_version (tmp_dir, arch_patch_id_branch (&patch_id));

      arch_patch_id_finalise (&patch_id);
    }

    tree = arch_project_tree_new (tmp_dir);

    index = arch_source_inventory (tree, 1, 0, 0);
    index_by_name_path = file_name_in_vicinity (0, tmp_dir, ",,index-by-name");
    index_path = file_name_in_vicinity (0, tmp_dir, ",,index");

    rmrf_file (index_by_name_path);
    rmrf_file (index_path);

    index_by_name_fd = safe_open (index_by_name_path, O_WRONLY | O_CREAT | O_EXCL, 0444);
    rel_print_pika_escape_iso8859_1_table (index_by_name_fd, arch_escape_classes, index);
    safe_close (index_by_name_fd);

    rel_sort_table_by_field (0, index, 1);
    index_fd = safe_open (index_path, O_WRONLY | O_CREAT | O_EXCL, 0444);
    rel_print_pika_escape_iso8859_1_table (index_fd, arch_escape_classes, index);
    safe_close (index_fd);

    arch_snap_inode_sig (tree, archive, (t_uchar *)revision);

    rel_free_table (index);
    lim_free (0, index_by_name_path);
    lim_free (0, index_path);
    arch_project_tree_delete (tree);
  }

  safe_rename (tmp_dir, dest_directory);
}


/**
 * \brief Produce a revision, using any available resources
 * \param n_patches The number of patches that were applied is written here
 * \param dest_dir The directory to store the revision in.  It must not exist,
 * but its parent should.
 * \param arch (optional) an archive that may be useful for building the
 * revision
 * \param cache_dir A directory to look for pristine trees in
 * \param try_link If true, attempt to hard-link to a revision library
 * \param non_sparse_lib If non-NULL, a revision library to add intermediate
 * revisions to
 * \return 0 if no revision was built (but usually panics first) 
 * \return 1 if built by copying a revision library
 * \return 2 if built by hard-linking from a revision library
 * \return 3 if built from a pristine
 * \return 4 if built from an import revision
 * \return 5 if built from a cached revision
 *
 * The selection code is built on the following false assumptions:
 * 1 all patches are the same size
 * 2 cached trees take around as long to download as 16 patches
 * 3 throughput is the same for all archives
 * 
 * This is because we don't know the truth.  But
 * 1 on *average*, this is true (by definition).
 * 2 value reached by experimentation w/ a particular tree
 * 3 this a common case.
 */
static int
build_revision (int * n_patches,
                t_uchar const * const dest_dir,
                struct arch_archive * arch,
                t_uchar const * const archive, 
                t_uchar const * const revision,
                t_uchar const * const cache_dir,
                int try_link,
                t_uchar const * const non_sparse_lib)
{
  struct arch_archive * my_arch = arch;
  t_uchar const * my_archive = archive;
  t_uchar const * my_revision = revision;
  int revision_found = 0;
  int from_library = 0;
  int archive_revision_cost = 50;
  assoc_table contin_list = 0;

  rel_table loc_patches = 0;
  rel_table arch_patches = 0;
  rel_table patch_list = 0;
  int status = 0;

  find_patchlists (&loc_patches, &arch_patches, archive, revision, arch, &contin_list, cache_dir);

  if (rel_n_records(loc_patches)>0 || rel_n_records(arch_patches) >0)
    revision_found = 1;
  else
    {
      panic ("The requested revision cannot be built.  It is based on archives that are not registered.");
    }
  
  if (revision_found)
    {
      int n_arch_patches;
      patch_list = loc_patches;
      *n_patches = rel_n_records (patch_list);
      n_arch_patches = rel_n_records (arch_patches);
      if (*n_patches == 0 || (n_arch_patches > 0 && (*n_patches > n_arch_patches + archive_revision_cost)))
	{
	  patch_list = arch_patches;
	  rel_free_table (loc_patches);
	  *n_patches = rel_n_records(patch_list);
	}
      else
	{
	  rel_free_table (arch_patches);
	}

      if (patch_direction(patch_list, archive, revision) == -1)
        {
          my_archive = (patch_list)[*n_patches -1][0];
          my_revision = (patch_list)[*n_patches -1][1];
        }
      else 
        {
          my_archive = (patch_list)[0][0];
          my_revision = (patch_list)[0][1];
        }
      if (patch_list == arch_patches)
	{
	  struct arch_archive * tmp_arch = arch;
	  if (str_cmp (arch->official_name, my_archive))
            {
	      tmp_arch = arch_archive_connect_branch (my_archive, NULL);
              invariant (!!tmp_arch);
            }
	      
	  status = copy_archive_revision (dest_dir, tmp_arch, my_revision);
	  if (tmp_arch != arch)
	    arch_archive_close (tmp_arch);
	}
      else
	{
	  status = copy_local_revision (dest_dir, my_archive, my_revision, cache_dir, try_link);
	  invariant (status != 0);
	  from_library = (status == 2);
	}
    }
  else
    {
      make_patch_list(&patch_list, my_archive, my_revision, archive, revision, contin_list, arch);
    }
  if (my_arch != arch)
    arch_archive_close (my_arch);

  apply_revisions (dest_dir, arch, patch_list, patch_direction (patch_list, archive, revision), non_sparse_lib);
  rel_free_table (patch_list);
    {
      arch_patch_id patch_id;
      arch_project_tree_t * tree;
      
      arch_patch_id_init_archive (&patch_id, archive, revision);
      arch_set_tree_version (dest_dir, arch_patch_id_branch (&patch_id));

      tree = arch_project_tree_new (dest_dir);
      arch_snap_inode_sig (tree, archive, revision);

      arch_project_tree_delete (tree);
      arch_patch_id_finalise (&patch_id);
    }
  
  return status;
}

static void
select_patch_list (rel_table * patch_list, rel_table candidate)
{
  int candidate_n = rel_n_records (candidate);
  int current_n = rel_n_records (*patch_list);
  if (candidate_n < 1) 
    {
      rel_free_table (candidate);
      return;
    }
  else if (current_n < 1)
    {
      rel_free_table (*patch_list);
      * patch_list = candidate;
    }
  else if (current_n > candidate_n)
    {
      rel_free_table (*patch_list);
      * patch_list = candidate;
    }
}

static void 
make_patch_list (rel_table * patch_list, 
		 t_uchar const * const ancestor_archive, 
		 t_uchar const * const ancestor_revision, 
                 t_uchar const * const descendant_archive, 
		 t_uchar const * const descendant_revision, 
		 assoc_table contin_list,
		 struct arch_archive * arch)
{
  t_uchar * my_archive = str_save (0, descendant_archive);
  t_uchar * my_revision = str_save (0, descendant_revision);
  struct arch_archive * my_arch = arch;
  int entries = 0;

  

  while (str_cmp (my_archive, ancestor_archive) || str_cmp (my_revision, ancestor_revision))
    {
      rel_add_records (patch_list, rel_make_record (my_archive, my_revision, 0), 0);
      if (decrement_revision (&my_arch, &my_archive, &my_revision, contin_list, arch, descendant_archive, descendant_revision) == -1)
      {
	 rel_free_table (*patch_list);
	 *patch_list = 0;
	 if (my_archive != descendant_archive)
	    lim_free (0, my_archive);
      }
      
    }
  rel_add_records (patch_list, rel_make_record (my_archive, my_revision, 0), 0);
  lim_free (0, my_archive);
  lim_free (0, my_revision);
  if (my_arch != arch)
     arch_archive_close (my_arch);
  entries = rel_n_records(*patch_list);
  rel_reverse_table (*patch_list);
}


static int 
patch_direction (rel_table patch_list, t_uchar const * const archive, t_uchar const * const revision)
{
  int apply_direction = 1;
  if (rel_n_records(patch_list) > 0)
  {
    if (!str_cmp(archive, patch_list[0][0]) && !str_cmp(patch_list[0][1], revision))
      apply_direction = -1;
  }
  return apply_direction;
}

static int 
decrement_revision (struct arch_archive ** my_arch, 
                    t_uchar ** my_archive, 
                    t_uchar ** my_revision,
		    assoc_table contin_list,
                    struct arch_archive * arch, 
                    t_uchar const * const archive, 
                    t_uchar const * const revision)
{
  struct arch_archive *prev_arch = 0;
  t_uchar *prev_archive = 0;
  t_uchar *prev_revision = 0;

  if (find_for_previous_revision (&prev_arch, &prev_archive, &prev_revision, contin_list, *my_arch, *my_archive, *my_revision) == -1)
    return -1;

  if (*my_arch != prev_arch)
    {
      if (arch != *my_arch)
	arch_archive_close (*my_arch);
      *my_arch = prev_arch;
    }
  if (*my_archive != prev_archive)
    {
      if (archive != *my_archive)
	lim_free (0, *my_archive);
      *my_archive = prev_archive;
    }
  if (*my_revision != prev_revision)
    {
      if (revision != *my_revision)
	lim_free (0, *my_revision);
      *my_revision = prev_revision;
    }
  return 0;
}

static void
split_fq_name (t_uchar ** archive, t_uchar ** revision, t_uchar * fq_revision)
{
  *archive = arch_parse_package_name (arch_ret_archive, "", fq_revision);
  *revision =  arch_parse_package_name (arch_ret_non_archive, 0 , fq_revision);
}

static void
find_patchlists (rel_table * loc_patches,
		 rel_table * arch_patches,
		 t_uchar const * const archive, 
	 	 t_uchar const * const revision,
		 struct arch_archive * arch,
		 assoc_table * contin_list,
		 t_uchar const * const cache_dir)
{
  t_uchar * lib_descendant_archive = 0;
  t_uchar * lib_descendant_revision = 0;

  rel_table loc_ancestor_patches = 0;
  rel_table descendant_patches = 0;
  t_uchar * fq_loc_ancestor = 0;
  t_uchar * fq_arch_ancestor = 0;
  debug (dbg_builder, 1, "* Scanning for full-tree revision: ");
  find_prev_revision (contin_list, &fq_loc_ancestor, &fq_arch_ancestor, arch, revision, cache_dir);
  if (fq_loc_ancestor)
    {
      t_uchar * loc_ancestor_archive = arch_parse_package_name (arch_ret_archive, "", fq_loc_ancestor);
      t_uchar * loc_ancestor_revision =  arch_parse_package_name (arch_ret_non_archive, 0 , fq_loc_ancestor);
      make_patch_list (&loc_ancestor_patches, loc_ancestor_archive, loc_ancestor_revision, archive, revision, *contin_list, arch);
      lim_free (0, loc_ancestor_archive);
      lim_free (0, loc_ancestor_revision);
    }
  if (fq_arch_ancestor)
    {
      t_uchar * arch_archive, * arch_revision;
      split_fq_name (&arch_archive, &arch_revision, fq_arch_ancestor);
      make_patch_list (arch_patches, arch_archive, arch_revision, archive, revision, *contin_list, arch);
      lim_free (0, arch_archive);
      lim_free (0, arch_revision);
    }
  if (find_library_forwards (&lib_descendant_archive, &lib_descendant_revision, archive, revision))
    {
      if (direct_descendant (arch, revision, 
                             lib_descendant_revision))
	make_patch_list (&descendant_patches, archive, revision, lib_descendant_archive, lib_descendant_revision, *contin_list, arch);
    }
  debug (dbg_builder, 1, "\n");
  
  select_patch_list (loc_patches, descendant_patches);
  select_patch_list (loc_patches, loc_ancestor_patches);
}

/**
 * \brief Determine whether a revision is a descedant of a candidate revision
 * \note Because of in-version tags or imports, a namespace-subsequent revision
 * may not be a descendant.
 * \param arch An archive that may be useful.  May be NULL
 * \param revision The possible ancestor
 * \param candidate_descendant The possible descendant
 */
static int
direct_descendant (struct arch_archive * arch, 
                   t_uchar const * const revision, 
                   t_uchar const * const candidate_descendant)
{
  enum arch_revision_type type;
  int num_scanned;
  int answer;
  t_uchar * last_revision = scan_archive (&type, &num_scanned, NULL, NULL, arch, candidate_descendant, revision, -1);
  answer = (str_cmp (last_revision, revision) == 0);

  lim_free (0, last_revision);
  return answer;
}


/*
 * Note that tag-based branches, the previous revision is not an *ancestor*
 */
static t_uchar * 
find_previous_library (t_uchar const * const archive, t_uchar const * const revision)
{
  int r;
  int answer = 0;
  t_uchar * my_patchlevel = arch_parse_package_name (arch_ret_patch_level, "", revision);
  t_uchar * my_version = arch_parse_package_name (arch_ret_package_version, "", revision);
  rel_table revisions = arch_library_revisions (archive, my_version, 0);
  t_uchar * found_revision = 0;
  

  for (r = r = rel_n_records (revisions)-1; r >=0;  r--)
    {
      if (arch_cmp_revision (revisions[r][0], my_patchlevel) <= 0)
	{
	  answer=1;
	  break;
	}
    }
  if (answer)
    {
      found_revision = str_alloc_cat_many (0, my_version, "--", revisions[r][0], str_end);
      debug (dbg_builder, 3, "* found previous library revision %s/%s\n", archive, found_revision);
    }

  lim_free(0, my_version);
  lim_free(0, my_patchlevel);
  rel_free_table(revisions);

  return found_revision;
}


/*
 * Note that tag-based branches, the previous revision is not an *ancestor*
 */
static t_uchar *
find_previous_pristine (struct arch_archive *arch, t_uchar const * const revision, t_uchar const * const cache_dir)
{
  t_uchar * my_revision = 0;
  t_uchar * answer = 0;
  if (!cache_dir) return answer;
  my_revision = str_save (0, revision);
  while (my_revision)
    {
      t_uchar * prev_revision = 0;
      /****************************************************************
       * Try the pristine cache.
       */
      arch_project_tree_t * pristine_copy = arch_find_pristine (0, cache_dir, arch->official_name, my_revision, arch_any_pristine, arch_cache_dir_pristine_search);
      if (pristine_copy)
	{
	  arch_project_tree_delete (pristine_copy);
	  answer=str_save (0, my_revision);
	  break;
	}

      /********************************
       * Find the previous revision
       */
      prev_revision = arch_previous_revision (arch, my_revision);
      lim_free (0, my_revision);
      my_revision = prev_revision;
    }

  lim_free (0, my_revision);
  return answer;
}


/**
 * \brief Finds a ancestor or descendant full trees for a revision
 * \param lib_rev library revisions
 */
static int
find_multisource (t_uchar ** lib_rev, 
		  t_uchar ** arch_rev, 
		  t_uchar ** continuation, 
		  struct arch_archive * arch, 
		  t_uchar const * const revision,
		  t_uchar const * const cache_dir)
{
  int i = 0;
  t_uchar * candidate_revision = 0;
  t_uchar * pristine_candidate = 0;
  t_uchar * last_revision = 0;
  t_uchar * newest_cached = 0;
  t_uchar ** newest_cached_ptr = (arch_rev != NULL) ? &newest_cached : NULL;
  enum arch_revision_type type;
  debug (dbg_builder, 3, "* checking for %s/%s or earlier\n", arch->official_name, revision); 

  if (arch_rev && *arch_rev)
    arch_rev = 0;
  candidate_revision = find_previous_library (arch->official_name, revision);
  pristine_candidate = find_previous_pristine (arch, revision, cache_dir);
  if (pristine_candidate)
  {
    int use_pristine = 0;
    debug (dbg_builder, 3, "* found candidate pristine %s/%s\n", arch->official_name, pristine_candidate); 

    if (!candidate_revision)
      use_pristine = 1;
    else
      {
	t_uchar * candidate_level = arch_parse_package_name (arch_ret_patch_level, "", candidate_revision);
	t_uchar * pristine_level = arch_parse_package_name (arch_ret_patch_level, "", pristine_candidate);
	if (arch_cmp_revision (candidate_level, pristine_level) < 0)
	  use_pristine = 1;
      }

    if (use_pristine)
      {
	lim_free (0, candidate_revision);
	candidate_revision = pristine_candidate;
      }
    else
      lim_free (0, pristine_candidate);
  }

  last_revision = scan_archive (&type, &i, newest_cached_ptr, NULL,
                                arch, revision, candidate_revision, 50);
  if (arch_rev)
    {
      if (newest_cached)
        {
          *arch_rev = newest_cached;
          newest_cached = 0;
        }
      else if (type == arch_import_revision)
        {
          *arch_rev = str_save (0, last_revision);
        }
    }
  if (arch_rev && *arch_rev)
    {
      debug (dbg_builder, 3, "* found archive revision %s/%s\n", arch->official_name, *arch_rev); 
    }

  if (str_cmp (candidate_revision, last_revision) == 0)
    *lib_rev = candidate_revision;
  else
    {
      lim_free (0, candidate_revision);
      lib_rev = 0;
    }

  if (type == arch_continuation_revision)
    {
      *continuation = str_save (0, last_revision);
    }

  lim_free (0, last_revision);
  lim_free (0, newest_cached);
  return i;
}


static void 
find_prev_revision (assoc_table  * contin_list,
		    t_uchar ** local_rev, 
		    t_uchar ** arch_rev, 
		    struct arch_archive * arch, 
		    t_uchar const * const start_revision,
		    t_uchar const * const cache_dir)
{
  struct arch_archive * my_arch = arch;
  int patch_n = 0;
  int arch_rev_n = -1;
  t_uchar * my_start_revision = str_save (0, start_revision);
  *arch_rev = 0; 
  while (1)
    {
      t_uchar * fq_prev_revision = 0;
      t_uchar * continuation = 0;
      patch_n += find_multisource (local_rev, (*arch_rev == 0)?arch_rev : 0, &continuation, my_arch, my_start_revision, cache_dir);
      lim_free (0, my_start_revision);
      my_start_revision = 0;
      if (*local_rev)
	{
	  *local_rev = full_revision (my_arch->official_name, *local_rev);
	}
      /* BUG FOR MULTIPLE CACHEREVS */
      if (*arch_rev && arch_rev_n == -1)
      {
	rel_table arch_rev_patches = 0;
	make_patch_list (&arch_rev_patches, my_arch->official_name, 
	  *arch_rev, arch->official_name, start_revision, *contin_list, arch);
	arch_rev_n = rel_n_records (arch_rev_patches) - 1;
	*arch_rev = full_revision (my_arch->official_name, *arch_rev);
      }
      if (continuation)
	{
	  t_uchar * fq_key = full_revision (my_arch->official_name, continuation);
	  fq_prev_revision = arch_get_continuation (my_arch, continuation);
	  if (!arch_valid_package_name (fq_prev_revision, arch_req_archive, arch_req_patch_level, 0))
	    {
	      safe_printfmt (2, "arch_build_revision: archive contains bogus CONTINUATION file\n  archive: %s\n  revision: %s\n bogosity: %s\n",
			     my_arch->official_name, continuation, fq_prev_revision);
	      exit (2);
	    }
	  assoc_set (contin_list, fq_key, fq_prev_revision );
	  lim_free (0, fq_prev_revision);
	  lim_free (0, continuation);
	  fq_prev_revision = assoc_ref (*contin_list, fq_key);
	}

      if (*local_rev)
	  return;

      if (arch_rev_n >= 0 && arch_rev_n + 50 < patch_n)
	return;

      if (fq_prev_revision)
	{
	  t_uchar * prev_archive = arch_parse_package_name (arch_ret_archive, "", fq_prev_revision);
	  my_start_revision = arch_parse_package_name (arch_ret_non_archive, 0, fq_prev_revision);
	  if ( str_cmp (my_arch->official_name, prev_archive) )
	    {
	      if (my_arch != arch)
		arch_archive_close (my_arch);
	      my_arch = arch_archive_connect_readonly (prev_archive);
	    }
	  if (!my_arch)
	    debug (dbg_builder, 3, "* archive %s is not registered\n", prev_archive);
	  lim_free (0, prev_archive);
	}
      if (!my_arch || !my_start_revision)
	break;
    }
  if (my_arch != arch)
    arch_archive_close (my_arch);
}


static t_uchar *
scan_archive(enum arch_revision_type * type,
             int * num_scanned,
             t_uchar ** newest_cached, 
             t_uchar ** oldest_cached, 
	     struct arch_archive *arch, 
	     t_uchar const * const start_revision, 
	     t_uchar const * const end_revision,
             int max_count_after_cacherev)
{
  int count_after_cacherev = 0;
  t_uchar * my_revision = str_save (0, start_revision); 
  debug (dbg_builder, 6, "scanning from '%s'\n", my_revision);

  if (newest_cached != NULL)
    *newest_cached = NULL;
  if (oldest_cached != NULL)
    *oldest_cached = NULL;

  for (*num_scanned = 0; count_after_cacherev < max_count_after_cacherev || max_count_after_cacherev == -1; ++*num_scanned)
    {
      int is_cached = 0;
      int * is_cached_ptr = &is_cached;
      t_uchar * prev_revision = 0;
      if ((newest_cached == NULL || *newest_cached != NULL) && oldest_cached == NULL)
        {
          is_cached_ptr = NULL;
          count_after_cacherev ++;
        }
      
      arch_revision_type (type, is_cached_ptr, NULL, arch, my_revision);
      debug (dbg_builder, 1, ".");
      if (is_cached)
        {
          if (newest_cached != NULL && *newest_cached == 0)
            *newest_cached = str_save (0, my_revision);
          if (oldest_cached != NULL)
            {
              lim_free (0, *oldest_cached);
              *oldest_cached = str_save (0, my_revision);
            }
        }
      if (*type != arch_simple_revision)
        break;
      
      if (str_cmp (my_revision, end_revision) == 0)
        break;
      prev_revision = arch_previous_revision (arch, my_revision);
      /* simple revisions are always in the same archive namespace */
        {
          t_uchar * current_fqrevision = arch_fully_qualify (arch->official_name, my_revision);
          t_uchar * previous_fqrevision = arch_fully_qualify (arch->official_name, prev_revision);
          arch_ancestry_note_ancestor (current_fqrevision, previous_fqrevision);
          lim_free (0, current_fqrevision);
          lim_free (0, previous_fqrevision);
        }
      lim_free (0, my_revision);
      my_revision = prev_revision;
    }
  return my_revision;
}

int
find_library_forwards (t_uchar ** found_archive, t_uchar ** found_revision, t_uchar const * const archive, t_uchar const * const revision)
{
  int r;
  int answer = 0;
  t_uchar * my_patchlevel = arch_parse_package_name (arch_ret_patch_level, "", revision);
  t_uchar * my_version = arch_parse_package_name (arch_ret_package_version, "", revision);
  rel_table revisions = arch_library_revisions (archive, my_version, 0);
  
  debug (dbg_builder, 3, "* checking libraries for  %s/%s--%s or later\n", archive, my_version, my_patchlevel); 
  for (r = 0; r < rel_n_records (revisions);  r++ )
    {
      if (arch_cmp_revision (revisions[r][0], my_patchlevel) >= 0)
	{
	  answer=1;
	  break;
	}
    }
  if (answer)
    {
      *found_archive = str_save (0, archive);
      *found_revision = str_alloc_cat_many (0, my_version, "--", revisions[r][0], str_end);
      debug (dbg_builder, 3, "* found descendant revision %s/%s\n", *found_archive, *found_revision);
    }

  lim_free(0, my_version);
  lim_free(0, my_patchlevel);
  rel_free_table(revisions);

  return answer;
}


static int 
copy_local_revision (t_uchar const * const dest_dir,  
		     t_uchar const * const archive, 
		     t_uchar const * const revision, 
		     t_uchar const * const cache_dir,
		     int try_link)
{
  if (copy_pristine_revision (dest_dir, archive, revision, cache_dir))
    return 3;
  else 
    return copy_library_revision (dest_dir, archive, revision, try_link);
  return 0; 
}


static int 
copy_library_revision (t_uchar const * const dest_dir, 
	      t_uchar const * const archive,  
	      t_uchar const * const revision,
	      int try_link)
{
  rel_table index_by_name = 0;
  int status = 0;
  arch_patch_id patch_id;
  arch_project_tree_t * library_rev_tree;
  
  arch_patch_id_init_archive (&patch_id, archive, revision);
  
  library_rev_tree = arch_library_find (0, &patch_id, 1);
  if (!library_rev_tree) return status;
  if (try_link)
    {
      struct stat from, to;
      safe_stat (library_rev_tree->root, &from);
      safe_stat (dest_dir, &to);
      if (from.st_dev == to.st_dev)
	{
	  debug (dbg_builder, 1, "* from revision library (linking): %s/%s\n", archive, revision);
	  safe_rmdir (dest_dir);
	  build_link_tree (library_rev_tree->root, dest_dir);
	  status = 2;
	  goto link_done;
	}
    }
  debug (dbg_builder, 1, "* from revision library: %s/%s\n", archive, revision);

  index_by_name = arch_library_index (archive, revision);
  rel_sort_table_by_field (0, index_by_name, 0);
  copy_file_list (dest_dir, library_rev_tree->root, index_by_name);

  rel_free_table (index_by_name);
  status = 1;

  link_done:
  arch_project_tree_delete (library_rev_tree);
  arch_patch_id_finalise (&patch_id);
  return status;
}


static int 
copy_pristine_revision (t_uchar const * const dest_dir,  
	       t_uchar const * const archive, 
	       t_uchar const * const revision, 
	       t_uchar const * const cache_dir)
{
  arch_project_tree_t * pristine_copy;
  if (!cache_dir) return 0;
  pristine_copy = arch_find_pristine (0, cache_dir, archive, revision, arch_any_pristine, arch_cache_dir_pristine_search);

  if (!pristine_copy)
      return 0;

  rel_table file_list = 0;

  debug (dbg_builder, 1, "* from pristine cache: %s/%s\n", archive, revision);

  file_list = arch_source_inventory (pristine_copy, 1, 0, 0);
  copy_file_list (dest_dir, pristine_copy->root, file_list);

  rel_free_table (file_list);
  arch_project_tree_delete(pristine_copy);

  return 1;
}

static void
move_tmp_revision (t_uchar const * const dest_dir, t_uchar const * const tmpdir)
{
  int x;
  rel_table files = directory_files (tmpdir);

  for (x = 0; x < rel_n_records (files); ++x)
    {
      if (str_cmp (".", files[x][0]) && str_cmp ("..", files[x][0]))
	{
	  t_uchar * frm = 0;
	  t_uchar * to = 0;

	  frm = file_name_in_vicinity (0, tmpdir, files[x][0]);
	  to = file_name_in_vicinity (0, dest_dir, files[x][0]);

	  safe_rename (frm, to);

	  lim_free (0, frm);
	  lim_free (0, to);
	}
    }

  safe_rmdir (tmpdir);
  rel_free_table (files);
}


static int 
copy_archive_revision(t_uchar const * const dest_dir,
		      struct arch_archive *arch,
		      t_uchar const * const revision)
{
  enum arch_revision_type type;
  int is_cached;
  arch_revision_type (&type, &is_cached, NULL, arch, revision);
  
  /********************************
   * Archive Cached Revision?
   */
  if (is_cached)
  {
    copy_cached_revision (dest_dir, arch, revision);
    /* ensure that inode sigs from the archive are ignored */
    arch_clear_inode_sigs (dest_dir);
    return 5;
  }

  /********************************
   * Import Revision?
   */
  else if (type == arch_import_revision)
  {
    copy_import_revision (dest_dir, arch, revision);
    /* ensure that inode sigs from the archive are ignored */
    arch_clear_inode_sigs (dest_dir);
    return 4;
  }
  
  else return 0;
}


static void
copy_cached_revision(t_uchar const * const dest_dir, 
		     struct arch_archive * arch, 
		     t_uchar const * const revision)
{
  t_uchar * tmpdir = 0;
  debug (dbg_builder, 1, "* from archive cached: %s/%s\n", arch->official_name, revision);

  tmpdir = tmp_file_name (dest_dir, ",,cached-revision");
  arch_get_cached_revision (arch, revision, tmpdir);
  move_tmp_revision (dest_dir, tmpdir);
  lim_free (0, tmpdir);
}


static void 
copy_import_revision (t_uchar const * const dest_dir, 
		      struct arch_archive * arch, 
		      t_uchar const * const revision)
{
  t_uchar * tmpdir = 0;

  debug (dbg_builder, 1, "* from import revision: %s/%s\n", arch->official_name, revision);

  tmpdir = tmp_file_name (dest_dir, ",,import-revision");
  arch_get_import_revision (arch, revision, tmpdir);
  move_tmp_revision (dest_dir, tmpdir);
  lim_free (0, tmpdir);
}

static t_uchar * 
full_revision (t_uchar const * const archive, t_uchar const * const revision)
{
  return str_alloc_cat_many (0, archive, "/", revision, str_end);
}

static int
find_for_previous_revision (struct arch_archive ** prev_arch, 
                            t_uchar **prev_archive, 
			    t_uchar **prev_revision, 
			    assoc_table contin_list,
			    struct arch_archive * arch, 
			    t_uchar *archive,
			    t_uchar *revision)
{
  t_uchar * fq_rev = full_revision (arch->official_name, revision);
  t_uchar * fq_prev_revision = assoc_ref (contin_list, fq_rev);
  lim_free (0, fq_rev);
  if (!fq_prev_revision)
    {
      *prev_revision = arch_previous_revision (arch, revision);
      if (! *prev_revision)
	return -1;
      *prev_arch = arch;
      *prev_archive = str_save (0, archive);
      lim_free (0, fq_prev_revision);
    }
  else
    {
      if (!arch_valid_package_name (fq_prev_revision, arch_req_archive, arch_req_patch_level, 0))
	{
	  safe_printfmt (2, "arch_build_revision: continuation list contains bogus entry\n  archive: %s\n  revision: %s\n bogosity: %s\n",
			 archive, revision, fq_prev_revision);
	  exit (2);
	}
      *prev_archive = arch_parse_package_name (arch_ret_archive, 0, fq_prev_revision);
      *prev_revision = arch_parse_package_name (arch_ret_non_archive, 0, fq_prev_revision);
      
      *prev_arch = arch_archive_connect_readonly (*prev_archive);
      if ( *prev_arch == 0)
	return -1;
      
    }
  return 0;
}


static void
apply_revision_patch (t_uchar const * const dest_tree_root, 
		      struct arch_archive * arch, 
		      t_uchar const * const revision,
		      int reverse)
{
  t_uchar * tmppatch = 0;
  struct arch_apply_changeset_report r;
  arch_project_tree_t * dest_tree;
  
  dest_tree = arch_project_tree_new (dest_tree_root);
  tmppatch = tmp_file_name (dest_tree->root, ",,next-patch");
  arch_get_patch (arch, revision, tmppatch);

  mem_set0 ((t_uchar *)&r, sizeof (r));

  arch_apply_changeset (&r, tmppatch, dest_tree, arch_unspecified_id_tagging, arch_inventory_unrecognized, reverse, 0, 0);

  if (arch_conflicts_occurred (&r))
    panic ("conflict applying patch in arch_build_revision");

  arch_free_apply_changeset_report_data (&r);

  rmrf_file (tmppatch);
  lim_free (0, tmppatch);
  arch_project_tree_delete (dest_tree);
}


/**
 * \brief Apply the changesets for listed revisions to the specified directory
 * 
 * The specified directory must contain a tree which can be patched into
 * the right revision.  If going forwards, it must be the first-listed revision.
 * If going backwards, it must be the last-listed revision.
 *
 * The revision list is actually a list of deltas: for each revision n, the
 * delta between patch_list[n-1] and patchlist[n] will be applied.  However,
 * this implementation requires that n is a direct descendant of n-1.
 *
 * \param dest_dir Directory to apply changesets to
 * \param arch An archive which may be useful for this.  If it is not useful,
 * it will be ignored.  It may also be NULL.
 * \param patch_list The list of deltas to apply, archive in column 0,
 * revision in 1.
 * \param apply_direction +1 to apply from 0 to the end, -1 to apply from end
 * to 0
 * \param non_sparse_lib If non-NULL, a non-sparse library for revisions
 */
static void 
apply_revisions (t_uchar const * const dest_dir, 
		 struct arch_archive * arch, 
		 rel_table patch_list,
		 int apply_direction,
                 t_uchar const * non_sparse_lib)
{
  int i;
  struct arch_archive * my_arch = arch;
  int reverse = (apply_direction == 1) ? 0 : 1;
  int start;
  int end;
  t_uchar * reverse_text = (apply_direction < 0)? " (in reverse)" : "";

  if (apply_direction == 1)
  {
     start = 1;
     end = rel_n_records (patch_list);
  }
  else
  {
     start = rel_n_records(patch_list) -1;
     end = 0;
  }
  if (start != end)
    debug (dbg_builder, 1, "* Applying %d revisions%s: ", rel_n_records(patch_list) - 1, reverse_text);
  for (i=start; i != end; i +=apply_direction)
  {
    int target_rev = i;
    /* non-sparse adding should not be done for the final target revision */
    if (i+apply_direction == end)
      non_sparse_lib = NULL;
    if (reverse)
      target_rev--;
    if (my_arch == 0 || str_cmp(my_arch->official_name, patch_list[i][0]))
      {
	if (my_arch != arch)
	  arch_archive_close (my_arch);
	my_arch = arch_archive_connect_readonly (patch_list[i][0]);
        if (!my_arch)
          {
            safe_printfmt (2, "Could not connect to %s - unable to apply patch %s\n", patch_list[i][0], patch_list[i][1]);
            exit (2);
          }
      }

    debug (dbg_builder, 1, ".");    
    apply_revision_patch (dest_dir, my_arch, patch_list[i][1], reverse);
    if (non_sparse_lib != NULL)
      {
        add_to_library (dest_dir, non_sparse_lib, my_arch->official_name, patch_list[target_rev][1]);
      }
  }
  if (start != end)
    debug (dbg_builder, 1, "\n");

  if (my_arch != arch)
    arch_archive_close (my_arch);
}

/**
 * \brief Add a reference tree to the library.  
 *
 * The tree is hard-linked, not copied or moved.  Assumed to be on the same 
 * filesystem as the library.
 * \param tree_dir The directory containing the reference tree
 * \param library The library to add the the revision to
 * \param archive The name of the archive containing the revision
 * \param revision The unqualified revision name
 */
void add_to_library (t_uchar const * const tree_dir, 
                     t_uchar const * const library, 
                     t_uchar * archive, 
                     t_uchar * revision)
{
  t_uchar * tmp_lib = str_save (0, library);
  t_uchar * dest_directory = arch_library_revision_dir_in_lib (tmp_lib, archive, revision);
  t_uchar * dest_dir_dir = file_name_directory_file (0, dest_directory);
  t_uchar * tmp_dir = tmp_file_name (dest_dir_dir, ",,new-revision");  
  
  ensure_directory_exists (dest_dir_dir);
  build_link_tree(tree_dir, tmp_dir);
  finalize_library_revision (dest_directory, tmp_dir, NULL, archive, revision);

  lim_free (0, tmp_dir);
  lim_free (0, dest_dir_dir);
  lim_free (0, dest_directory);
  lim_free (0, tmp_lib);
}




/* tag: Tom Lord Wed May 21 15:59:22 2003 (build-revision.c)
 */
